/*
 * PROGRAM:
 * vtkcellgen
 *
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */
#include "globals.h"

int get_number_of_cells_per_volume(int celltype)
{
  if (celltype == VTK_TRIANGLE ||
      celltype == VTK_QUADRATIC_TRIANGLE ||
      celltype == VTK_WEDGE ||
      celltype == VTK_QUADRATIC_WEDGE ||
      celltype == VTK_QUADRATIC_LINEAR_WEDGE || 
      celltype == VTK_BIQUADRATIC_QUADRATIC_WEDGE)
    {
      return 2;
    }
  else if (celltype == VTK_CUBIC_EDGE)
    {
      return 8;
    }
  else if (celltype == VTK_TETRA)
    {
      return 6;
    }
  else
    {
      return 1;
    }

  return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int
get_number_of_points_of_cell (celltype)
{
  switch (celltype)
    {
    case VTK_POLY_VERTEX:
      return 1;
    case VTK_TRIANGLE:
      return 3;
    case VTK_TETRA:
    case VTK_QUAD:
    case VTK_CUBIC_EDGE:
      return 4;
    case VTK_HEXAHEDRON:
    case VTK_QUADRATIC_QUAD:
      return 8;
    case VTK_WEDGE:
    case VTK_QUADRATIC_TRIANGLE:
    case VTK_QUADRATIC_LINEAR_QUAD:
      return 6;
    case VTK_BIQUADRATIC_QUAD:
      return 9;
    case VTK_QUADRATIC_WEDGE:
      return 15;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      return 12;
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
      return 18;
    case VTK_QUADRATIC_HEXAHEDRON:
      return 20;
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
      return 24;
    case VTK_TRIQUADRATIC_HEXAHEDRON:
      return 27;
    case VTK_BICUBIC_QUAD:
      return 16;
    case VTK_TRICUBIC_HEXAHEDRON:
      return 64;
    }
  return 0;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
compute_node_numbers (parameters *params, int *xnum, int *ynum, int *znum)
{
   int cellx = params->cellx;
   int celly = params->celly;

  switch (params->celltype)
    {
    case VTK_POLY_VERTEX:
      *xnum = cellx;
      *ynum = celly;
      *znum = 1;
      break;
    case VTK_TRIANGLE:
    case VTK_QUAD:
      *xnum = cellx + 1;
      *ynum = celly + 1;
      *znum = 1;
      break;
    case VTK_TETRA:
    case VTK_WEDGE:
    case VTK_HEXAHEDRON:
      *xnum = cellx + 1;
      *ynum = celly + 1;
      *znum = 2;
      break;
    case VTK_QUADRATIC_LINEAR_QUAD:
      *xnum = cellx * 2 + 1;
      *ynum = celly + 1;
      *znum = 1;
    case VTK_QUADRATIC_QUAD:
    case VTK_QUADRATIC_TRIANGLE:
    case VTK_BIQUADRATIC_QUAD:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 1;
      break;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 2;
      break;
    case VTK_QUADRATIC_WEDGE:
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
    case VTK_TRIQUADRATIC_HEXAHEDRON:
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
    case VTK_QUADRATIC_HEXAHEDRON:
      *xnum = cellx * 2 + 1;
      *ynum = celly * 2 + 1;
      *znum = 3;
      break;
    case VTK_CUBIC_EDGE:
    case VTK_BICUBIC_QUAD:
      *xnum = cellx * 3 + 1;
      *ynum = celly * 3 + 1;
      *znum = 1;
      break;
    case VTK_TRICUBIC_HEXAHEDRON:
      *xnum = cellx * 3 + 1;
      *ynum = celly * 3 + 1;
      *znum = 4;
      break;
    default:
      printf ("ERROR:  cell type %i is not supported\n", params->celltype);
      exit (0);
    }

  return;
}

