/*
 * PROGRAM:
 * vtkcellgen
 *
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_cells (FILE * out, parameters *params)
{
  int i, j, nodenum;
  int xoff = 0, yoff = 0, zoff = 0;
  int xnum, ynum, znum;
  int cellx = params->cellx;
  int celly = params->celly;
  int celltype = params->celltype;
  int volcells;
  
  compute_node_numbers (params, &xnum, &ynum, &znum);
  nodenum = get_number_of_points_of_cell (celltype);
  volcells = get_number_of_cells_per_volume(celltype)*cellx*celly;

  if (params->quite == 0)
    {
      printf ("Creating cells\n");
    }
  
  fprintf (out, "CELLS %i %i\n", volcells, volcells * nodenum + volcells);
  
  if (celltype == VTK_POLY_VERTEX)
    {
      for (i = 0; i < cellx * celly; i++)
	{
	  fprintf (out, "%i ", nodenum);
	  for (j = 0; j < nodenum; j++)
	    {
	      fprintf (out, "%i ", i * nodenum + j);
	    }
	  fprintf (out, "\n");
	}
    }
  else
    {
      for (j = 0; j < celly; j++)
	{
	  for (i = 0; i < cellx; i++)
	    {
	      print_single_cell (out, xnum, ynum, znum, &xoff, &yoff, &zoff, celltype);
	    }
	}
    }

  fprintf (out, "CELL_TYPES %i \n", volcells);
  for (i = 0; i < volcells; i++)
    fprintf (out, "%i\n", celltype);
    
  return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int
print_single_cell (FILE * out, int cols, int rows, int depths, int *xoffset, int *yoffset, int *zoffset, int celltype)
{
  int retval = 1;
  int nodenum = get_number_of_points_of_cell (celltype);

  switch (celltype)
    {
   case VTK_TRIANGLE:
      /*first */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      /*second */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_TETRA:
      /*first*/
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + cols * rows);
      /*second*/
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + cols * rows);
      /*third*/
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + cols * rows);
      /*fourth*/
      fprintf (out, "%i ", nodenum);
      
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + cols * rows);
      /*fifth*/
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + cols * rows);
      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      /*sixth*/
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ",  cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ",  cols * (*yoffset + 1) + *xoffset + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);

           *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      /*second */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      *xoffset += 1;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_QUADRATIC_TRIANGLE:
      /*first */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      /*second */
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_LINEAR_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 1;
	}
      break;
    case VTK_QUADRATIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_LINEAR_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + cols * rows);

      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUADRATIC_WEDGE:
      /*first */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + cols * rows);
      /*face center nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + cols * rows);
      /*second */
      fprintf (out, "%i ", nodenum);
      /*lower triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      /*upper triangle */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      /*lower midege nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      /*midedge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + cols * rows);
      /*face center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_QUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      /*middle side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_TRIQUADRATIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);
      /*lower edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      /*upper edge center nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      /*middle quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      /*mid side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      /*vertical mid side nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 1 * cols * rows);
      *xoffset += 2;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 2;
	}
      break;
    case VTK_CUBIC_EDGE:
       /*I*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset) + *xoffset + 2);
       /*II*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 1) + *xoffset + 2);
       /*III*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2);
       /*IV*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 3) + *xoffset + 2);
      /*X-direction */
       /*I*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset);
       /*II*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1);
       /*III*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 2);
       /*IV*/ fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 3);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    case VTK_BICUBIC_QUAD:
      fprintf (out, "%i ", nodenum);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);

      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);

      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    case VTK_TRICUBIC_HEXAHEDRON:
      fprintf (out, "%i ", nodenum);
      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset);
      /*upper quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 * cols * rows);

      /*lower midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset);

      /*upper midedge nodes */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 * cols * rows);

      /*lower quad */
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 * cols * rows);

      /*face I */
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 * cols * rows);

      /*face II */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 3 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 3 + 2 * cols * rows);

      /*face III */
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 0) + *xoffset + 1 + 2 * cols * rows);

      /*face IV */
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 3) + *xoffset + 2 + 2 * cols * rows);

      /*face V */
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 0 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 0 * cols * rows);

      /*face VI */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 3 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 3 * cols * rows);

      /*volume center nodes */
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 1 + 1 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 1 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 1) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i ", cols * (*yoffset + 2) + *xoffset + 2 + 2 * cols * rows);
      fprintf (out, "%i\n", cols * (*yoffset + 2) + *xoffset + 1 + 2 * cols * rows);

      *xoffset += 3;
      if ((*xoffset) == (cols - 1))
	{
	  *xoffset = 0;
	  *yoffset += 3;
	}
      break;
    }

  return retval;
}

