/*
 * PROGRAM:
 * vtkcellgen
 *
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
ef_pointer get_elevfunction(parameters *params)
{
  if(strncmp(params->shape_func_name, "simple", 6) == 0)
    return elevfunc_simple;

  if(strncmp(params->shape_func_name, "parabel", 7) == 0)
    return elevfunc_parabel;

  if(strncmp(params->shape_func_name, "cosinus", 7) == 0)
    return elevfunc_cosinus;

  if(strncmp(params->shape_func_name, "sinus", 5) == 0)
    return elevfunc_sinus;

  return  elevfunc_standard;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_numbering(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
	return (double) (k * xnum * ynum) + (double) (j * xnum) + (double) (i);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_standard(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
	return dz * (double) (k);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_simple(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
	return dz * ((double) (k) + sin ((double) i / (double) (0.1 * xnum)) +
		       cos ((double) j / (0.1 * ynum)));
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_parabel(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
        double xrange = (double)(xnum*dx)/2.0;
        double yrange = (double)(ynum*dy)/2.0;

	return dz * ((double) (k) + (1.2/(xrange*yrange))* ((i*dx - xrange)*(i*dx - xrange)+(j*dy - yrange)*(j*dy - yrange)));
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_cosinus(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
        double xrange = (double)(xnum*dx)/2.0;
        double yrange = (double)(ynum*dy)/2.0;

	return dz * ((double) (k) +  -cos((1.2/(xrange*yrange))*3.1415926535*((i*dx - xrange)*(i*dx - xrange)+(j*dy - yrange)*(j*dy - yrange)))/2);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
double elevfunc_sinus(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz)
{
        double xrange = (double)(xnum*dx)/2.0;
        double yrange = (double)(ynum*dy)/2.0;

	return dz * ((double) (k) +  -sin((1.2/(xrange*yrange))*3.1415926535*((i*dx - xrange)*(i*dx - xrange)+(j*dy - yrange)*(j*dy - yrange)))/2);
}

