/*
 * PROGRAM:
 * vtkcellgen
 * 
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_help (void)
{
  fprintf (stderr,"vtkcellgen [OPTIONS] \n");
  fprintf (stderr,"Generate a plate (unstructured grid) of a specific VTK cell\n");
  fprintf (stderr,"\n");
  fprintf (stderr,"vtkcellgen is a very simple cell generator to support the generation\n");
  fprintf (stderr,"of a mesh build of linear, quadratic or cubic VTK cells.\n");
  fprintf (stderr,"Based on one cell type a plate (unstructured grid) will be created with a specific number of\n");
  fprintf (stderr,"cells in x and y direction. Scalar, vector and tensor data are generated as\n");
  fprintf (stderr,"point and cell data. The output is written as \n");
  fprintf (stderr,"VTK legacy ASCII format to stdout or a specified file.\n");
  fprintf (stderr,"\n");
  fprintf (stderr,"The following options are supported:          \n");
  fprintf (stderr,"  --help, -h, -help, help      show this help and exit\n");
  fprintf (stderr,"  -q                           quite\n");
  fprintf (stderr,"  --celltype=VTK_CELL_TYPE     set the cell type which should\n");
  fprintf (stderr,"                               be generated, default: %i\n", DEFAULT_CELL);
  fprintf (stderr,"  --cellx=integer              the number of cells in x direction, default: %i\n", CELLNUMX);
  fprintf (stderr,"  --celly=integer              the number of cells in y direction, default: %i\n", CELLNUMY);
  fprintf (stderr,"  --xlength=double             length in x direction, default: %g\n", XLENGTH);
  fprintf (stderr,"  --ylength=double             length in y direction, default: %g\n", YLENGTH);
  fprintf (stderr,"  --zlength=double             length in z direction, default: %g\n", ZLENGTH);
  fprintf (stderr,"  --output=filename            the name of the output file, eg: quads.vtk\n");
  fprintf (stderr,"                               default: stdout\n");
  fprintf (stderr,"  --shapefunc=function_name    set the a specific function to modify the x, y and z coordinates\n");
  fprintf (stderr,"                               default: flat plate is generated\n");
  fprintf (stderr,"\n");
  fprintf (stderr,"The following cell types are supported:     \n");
  fprintf (stderr,"                                            \n");
  fprintf (stderr,"    VTK_POLY_VERTEX                       2 \n");
  fprintf (stderr,"    VTK_TRIANGLE                          5\n");
  fprintf (stderr,"    VTK_QUAD                              9 \n");
  fprintf (stderr,"    VTK_TETRA                             10\n");
  fprintf (stderr,"    VTK_HEXAHEDRON                        12\n");
  fprintf (stderr,"    VTK_WEDGE                             13\n");
  fprintf (stderr,"    VTK_QUADRATIC_TRIANGLE                22\n");
  fprintf (stderr,"    VTK_QUADRATIC_QUAD                    23\n");
  fprintf (stderr,"    VTK_QUADRATIC_HEXAHEDRON              25\n");
  fprintf (stderr,"    VTK_QUADRATIC_WEDGE                   26\n");
  fprintf (stderr,"    VTK_BIQUADRATIC_QUAD                  28\n");
  fprintf (stderr,"    VTK_TRIQUADRATIC_HEXAHEDRON           29\n");
  fprintf (stderr,"    VTK_QUADRATIC_LINEAR_QUAD             30\n");
  fprintf (stderr,"    VTK_QUADRATIC_LINEAR_WEDGE            31\n");
  fprintf (stderr,"    VTK_BIQUADRATIC_QUADRATIC_WEDGE       32\n");
  fprintf (stderr,"    VTK_BIQUADRATIC_QUADRATIC_HEXAHEDRON  33\n");
  fprintf (stderr,"    VTK_CUBIC_EDGE                        34\n");
  fprintf (stderr,"    VTK_BICUBIC_QUAD                      36\n");
  fprintf (stderr,"    VTK_TRICUBIC_HEXAHEDRON               37\n");
  fprintf (stderr,"\n");
  fprintf (stderr,"The following function for x,y,z position calculation are supported:\n");
  fprintf (stderr,"    simple                     simple trigon-function x(x,y) = x\n");
  fprintf (stderr,"                                                      y(x,y) = y\n");
  fprintf (stderr,"                                                      z(x,y) = sin(x) + cos(y)\n");
  fprintf (stderr,"    cosinus                    nice cosinus function: x(x,y) = x\n");
  fprintf (stderr,"                                                      y(x,y) = y\n");
  fprintf (stderr,"                                                      z(x,y) = -cos(pi*((x)^2+(y)^2))/2\n");
  fprintf (stderr,"    sinus                      nice sinus function:   x(x,y) = x\n");
  fprintf (stderr,"                                                      y(x,y) = y\n");
  fprintf (stderr,"                                                      z(x,y) = -sin(pi*((x)^2+(y)^2))/2\n");
  fprintf (stderr,"    parabel                    parabel function:      x(x,y) = x\n");
  fprintf (stderr,"                                                      y(x,y) = y\n");
  fprintf (stderr,"                                                      z(x,y) = x^2 + y^2\n");
  fprintf (stderr,"    hyper                      hyperboloid function   x(x,y) = r*cos(x) - y*sin(x)\n");
  fprintf (stderr,"                                                      y(x,y) = r*sin(x) + y*cos(x)\n");
  fprintf (stderr,"                                                      z(x,y,z) = x*y*tan(210) + z\n");
  fprintf (stderr,"    screw                      screw function:        x(x,y) = (p - y)*cos(x)\n");
  fprintf (stderr,"                                                      y(x,y) = (p - y)*sin(x)\n");
  fprintf (stderr,"                                                      z(x,y,z) = p*x + z\n");
  fprintf (stderr,"\n");
}

