/*
 * PROGRAM:
 * vtkcellgen
 * 
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void
print_ugrid_points (FILE * out, parameters *params)
{
  int i, j, k;
  double x[3];
  double scale = 0;
  double xrange;
  double yrange;
  int xnum = params->xnum;
  int ynum = params->ynum;
  int znum = params->znum;
  double dx = params->dx;
  double dy = params->dy; 
  double dz = params->dz; 
  int zelev = params->flag_elev;
  sf_pointer func;
  
  xrange = (params->xlength)/2.0;
  yrange = (params->ylength)/2.0;

  func = get_shapefunction(params);

  if (params->quite == 0)
    {
      printf ("Creating points\n");
    }

  fprintf (out, "POINTS %i float\n", xnum * ynum * znum);

  if (zelev == 1)
    scale = 1;

  for (k = 0; k < znum; k++)
    for (j = 0; j < ynum; j++)
      for (i = 0; i < xnum; i++) 
        {
	func(xnum, ynum, znum, i + 0.5, j + 0.5, k + 0.5, dx, dy, dz, x);
	fprintf (out, "%g %g %g\n", x[0] - xrange, x[1] - yrange, x[2]);
	}

  return;
}

