/*
 * PROGRAM:
 * vtkcellgen
 * 
 * AUTHOR:
 * Soeren Gebbert
 * soerengebbert <at> googlemail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * The GNU General Public License is available here:
 * http://www.gnu.org/licenses/gpl.txt
 * 
 */


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "globals.h"

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
sf_pointer get_shapefunction(parameters *params)
{
  if(strncmp(params->shape_func_name, "simple", 6) == 0)
    return shapefunc_simple;

  if(strncmp(params->shape_func_name, "parabel", 7) == 0)
    return shapefunc_parabel;

  if(strncmp(params->shape_func_name, "cosinus", 7) == 0)
    return shapefunc_cosinus;

  if(strncmp(params->shape_func_name, "sinus", 5) == 0)
    return shapefunc_sinus;

  if(strncmp(params->shape_func_name, "hyper", 5) == 0)
    return shapefunc_hyperboloid;

  if(strncmp(params->shape_func_name, "screw", 5) == 0)
    return shapefunc_screw;

  return  shapefunc_standard;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_standard(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   x[0] = (double)i*dx;
   x[1] = (double)j*dy;
   x[2] = (double)k*dz;
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_simple(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   x[0] = (double)i*dx;
   x[1] = (double)j*dy;
   x[2] = elevfunc_simple(xnum, ynum, znum, i, j, k, dx, dy, dz);
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_parabel(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   x[0] = (double)i*dx;
   x[1] = (double)j*dy;
   x[2] = elevfunc_parabel(xnum, ynum, znum, i, j, k, dx, dy, dz);
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_cosinus(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   x[0] = (double)i*dx;
   x[1] = (double)j*dy;
   x[2] = elevfunc_cosinus(xnum, ynum, znum, i, j, k, dx, dy, dz);
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_sinus(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   x[0] = (double)i*dx;
   x[1] = (double)j*dy;
   x[2] = elevfunc_sinus(xnum, ynum, znum, i, j, k, dx, dy, dz);
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_hyperboloid(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   double a2, b2, c2;
   double x0, x1, x2;
   double r, alpha;


   a2 = 2*dx*dx*xnum*xnum;
   b2 = 2*dy*dy*ynum*ynum;
   c2 = 2*dz*dz*znum*znum;
    
   x0 = (double)i*dx;
   x1 = (double)j*dy;
   x2 = (double)k*dz;

   r = sqrt(x0*x0 + x1*x1);
   alpha = 210;
      
   x[0] = r*cos(x0) - x1*sin(x0);      
   x[1] = r*sin(x0) + x1*cos(x0);     
   x[2] = x0*x1*tan(alpha) + x2; 
    
   return;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void shapefunc_screw(int xnum, int ynum, int znum, int i, int j, int k, double dx, double dy, double dz, double *x)
{
   double x0, x1, x2;
   double r, p;

   x0 = (double)i*dx;
   x1 = (double)j*dy;
   x2 = (double)k*dz;

   r = sqrt(x0*x0 + x1*x1);
   p = 1;
      
   x[0] = (p - x1)*cos(x0);      
   x[1] = (p - x1)*sin(x0);     
   x[2] = p*x0 + x2; 
    
   return;
}

